﻿// BallPhysicsGame Version 3 - Using Papervision 2.0 Beta - Revision 876, JibLibFlash - Revision 33, Caurina Tweener - Revision 404
// A physics based ball game, where you must knock off red squares. Level information is loaded via an xml file
//   also makes use an advanced motion tracking spring camera
//
// released under MIT License
// http://www.opensource.org/licenses/mit-license.php
//
//@author		Devin Reimer
//@version		1.0.2
//@website 		http://www.permaven.com, http://blog.almostlogical.com


//Copyright (c) 2009 Devin Reimer

/*
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package com
{
	import com.objects.AdvancedPhysicsObject;
	import flash.display.DisplayObject;
	import flash.display.Shape;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.text.TextField;
	import flash.text.TextFormat;
	import flash.text.TextFormatAlign;
	import flash.ui.Keyboard;
	import flash.events.IOErrorEvent;
	import flash.display.Loader;
	import flash.net.URLLoader;
	import flash.net.URLRequest;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	import flash.geom.Point;
	import flash.utils.Timer;
	import org.ascollada.utils.FPS;
	
	import org.papervision3d.core.math.Number3D;
	import org.papervision3d.view.Viewport3D;
	import org.papervision3d.render.QuadrantRenderEngine;
	import org.papervision3d.materials.utils.MaterialsList;
	import org.papervision3d.materials.ColorMaterial;
	import org.papervision3d.materials.shadematerials.FlatShadeMaterial;
	import org.papervision3d.objects.primitives.Cube;
	import org.papervision3d.objects.primitives.Plane;
	import org.papervision3d.objects.primitives.Sphere;
	import org.papervision3d.objects.DisplayObject3D;
	import org.papervision3d.lights.PointLight3D;
	import org.papervision3d.scenes.Scene3D;

	import jiglib.math.*;
	import jiglib.geometry.*;
	import jiglib.physics.*;
	import jiglib.physics.constraint.*;
	
	import com.cameras.JigLibMotionSpringCamera3D;
	import com.LevelBuilder;
	import com.events.LevelLoadEvent;

	public class BallPhysicsGame extends Sprite
	{
		private static const PLAYER_BALL_RADIUS:int = 25;
		private static const PLAYER_FORCE:int = 125;
		private static const PLAYER_WEIGHT:int = 5;
		private static const PLAYER_COLOUR:uint = 0xFBA248;
		
		private static const MAX_SPEED:Number = 9;
		
		private static const BASIC_GROUND_RESTITUTION:Number = 0.5;
		private static const BASIC_GROUND_STATIC_FRICTION:Number = 100;
		
		private static const WORLD_BOTTOM:int = -300;
		private static const STARTING_LEVEL:int = 1;
		private static const LEVEL_XML_PATH:String = "xml/AdvancedLevelInfo.xml";
		//private static const LEVEL_XML_PATH:String = "workingExamples/3DBallAdventureV3/xml/AdvancedLevelInfo.xml";
		
		private var viewport:Viewport3D;
		private var scene:Scene3D;
		private var camera:JigLibMotionSpringCamera3D;
		private var renderer:QuadrantRenderEngine;
	
		private var lightSource:PointLight3D;
		
		private var currentLevel:int = STARTING_LEVEL;
		private var keyRight:Boolean = false;
		private var keyLeft:Boolean = false;
		private var keyForward:Boolean = false;
		private var keyBack:Boolean = false;
	
		private var xmlLoader:URLLoader
		
		private var playerBall:JSphere;
		
		private var remainingMarkers:int=1;
		private var totalMarkers:int = 1;
		private var startTime:Number;
		private var physicsObjArray:Array;
		
		private var currentSpeed:Number = 0;
		
		private var playerResetNumber3D:Number3D = new Number3D();
		private var playerResetRotationY:Number = 0;
		
		private var hud:HUD;
		private var levelBuilder:LevelBuilder;
		
		private var currentDate = new Date();
		
		
		public function BallPhysicsGame()
		{
			stage.addEventListener(MouseEvent.CLICK, remainingSetup, false, 0, true); //wait until the user click to start
			
			graphics.beginFill(0xC9F1ED);
			graphics.drawRect(0, 0, stage.stageWidth, stage.stageHeight);
			graphics.endFill();
			
			stage.mouseChildren = false;
			
			initPapervision3D();
			
			hud = new HUD(stage.stageWidth, stage.stageHeight);
			addChild(hud);
		}
		
		private function remainingSetup(evt:Event)
		{
			stage.removeEventListener(MouseEvent.CLICK, remainingSetup, false);
			stage.addEventListener( KeyboardEvent.KEY_DOWN, keyDownHandler );
			stage.addEventListener( KeyboardEvent.KEY_UP, keyUpHandler );
			
			hud.createHUD();
			createPlayer();
			
			camera.jigLibObjTarget = playerBall; //locks the camera onto the motion of this object
			
			levelBuilder = new LevelBuilder(lightSource);
			levelBuilder.addEventListener(LevelLoadEvent.LEVEL_LOAD_COMPLETE, levelLoadComplete, false, 0, true);
			levelBuilder.loadLevelXML(LEVEL_XML_PATH);
			
			/*
			var fps:FPS = new FPS();
			fps.y += 20;
			addChild(fps);
			*/
		}

		private function initPapervision3D():void
		{
			viewport = new Viewport3D(stage.stageWidth,stage.stageHeight);
			addChild(viewport);
			
			scene = new Scene3D();
			 
			lightSource = new PointLight3D(true, true);
			lightSource.x = 100;
			lightSource.y = 300;
			lightSource.z = -500;
					 			 
			camera = new JigLibMotionSpringCamera3D();

			camera.mass = 10;
			camera.damping = 10;
			camera.stiffness = 1;
				
			camera.lookOffset = new Number3D(0, 20, 30);
			camera.positionOffset = new Number3D(0, 100, -270);
			
			camera.focus = 100;
			camera.zoom = 10;
			 
			renderer = new QuadrantRenderEngine(QuadrantRenderEngine.CORRECT_Z_FILTER);
		}
		
		private function createPlayer():void
		{
			var ballMaterial = new FlatShadeMaterial(lightSource,PLAYER_COLOUR);
			var ballSkin:Sphere = new Sphere(ballMaterial,PLAYER_BALL_RADIUS);
			
			scene.addChild(ballSkin);
			playerBall = new JSphere(ballSkin, true, PLAYER_BALL_RADIUS);
			playerBall.setMass(PLAYER_WEIGHT);
			PhysicsSystem.getInstance().AddBody(playerBall);
		}
		
		private function levelLoadComplete(evt:LevelLoadEvent):void
		{
			physicsObjArray = levelBuilder.physicsObjArray;
			playerResetNumber3D = levelBuilder.playerResetNumber3D;
			playerResetRotationY = levelBuilder.playerResetRotationY;
			resetPlayerPosition();
			addPhysicsObject();
			startTime = new Date().getTime();
			beginRender();
		}
		
		private function addPhysicsObject():void
		{
			totalMarkers = 0;
			for each (var advancedPhysicsObj:AdvancedPhysicsObject in physicsObjArray)
			{
				PhysicsSystem.getInstance().AddBody(advancedPhysicsObj.physicsObj);
				scene.addChild(advancedPhysicsObj.physicsObj.BodySkin);
				if (advancedPhysicsObj.isMarker)
				{
					totalMarkers++;
				}
			}
				
			remainingMarkers = totalMarkers;
		}
	
		private function keyDownHandler(evt:KeyboardEvent):void
		{
			switch(evt.keyCode)
			{
				case Keyboard.UP:
					keyForward = true;
					keyBack = false;
					break;
				case Keyboard.DOWN:
					keyBack = true;
					keyForward = false;
					break;
				case Keyboard.LEFT:
					keyLeft = true;
					keyRight = false;
					break;
				case Keyboard.RIGHT:
					keyRight = true;
					keyLeft = false;
					break;
				case Keyboard.SPACE:
					trace(playerBall.CurrentState.Position.toString()); //for testing and positioning objects
					break;
			}
		}

		private function keyUpHandler(evt:KeyboardEvent):void
		{
			switch(evt.keyCode)
			{
				case Keyboard.UP:
					keyForward = false;
					break;
				case Keyboard.DOWN:
					keyBack = false;
					break;
				case Keyboard.LEFT:
					keyLeft = false;
					break;
				case Keyboard.RIGHT:
					keyRight = false;
					break;
			}
		}
		
		private function checkPlayerPosition():void
		{
			if (playerBall.CurrentState.Position.y < WORLD_BOTTOM) //if fall off level reset layer
			{
				resetPlayerPosition();
			}
		}
		
		private function resetPlayerPosition():void
		{			
			playerBall.MoveTo(new JNumber3D(playerResetNumber3D.x, playerResetNumber3D.y + PLAYER_BALL_RADIUS, playerResetNumber3D.z), JMatrix3D.IDENTITY);
			camera.target.rotationY = playerResetRotationY;  //resets rotationY of camera
		}

		private function beginRender():void
		{
			this.addEventListener(Event.ENTER_FRAME,render);
		}
		private function render(evt:Event):void
		{		
			calcPlayerForces();
			PhysicsSystem.getInstance().Integrate(0.1);
			checkPlayerPosition();
			checkAllPhysicsObjects();
			hud.updateHUDText(remainingMarkers, totalMarkers,new Date().getTime()-startTime);
			renderer.renderScene( scene, camera, viewport );
		}
		
		private function calcPlayerForces():void
		{
			var calcDisplayObj3D:DisplayObject3D = new DisplayObject3D(); //used to calculate positions
			var totalAvailableForce:Number = PLAYER_FORCE;
			var sidewaysForce:Number;
			var extraRotation:Number;
			calcDisplayObj3D.copyTransform(camera.target);
			calcDisplayObj3D.copyPosition(DisplayObject3D.ZERO);
			
			currentSpeed = getCurrentSpeed();
			
			if ((keyLeft && !keyRight) || (keyRight && !keyLeft)) //sideways force
			{
				sidewaysForce = PLAYER_FORCE / 2;
				totalAvailableForce -= sidewaysForce;
				
				if (keyLeft && !keyRight)//left
				{
					calcDisplayObj3D.moveLeft(sidewaysForce);
				}
				else //right
				{
					calcDisplayObj3D.moveRight(sidewaysForce);
				}
			}
			
			if (currentSpeed < MAX_SPEED)
			{
				if(keyForward && !keyBack)
				{
					calcDisplayObj3D.moveForward(totalAvailableForce);
				}
				else if(keyBack && !keyForward)
				{
					calcDisplayObj3D.moveBackward(totalAvailableForce);
				}
			}
			
			movePlayerBall(calcDisplayObj3D.x, calcDisplayObj3D.y, calcDisplayObj3D.z);
			
		}
		
		private function movePlayerBall(xForce:Number, yForce:Number, zForce:Number):void
		{
			var tractionRatio:Number = 1; //normal traction
			if (!hasTraction()) //if not touching object ball has no traction, so limit it movement
			{
				tractionRatio=3; //no traction
			}
			playerBall.AddWorldForce(new JNumber3D(xForce/tractionRatio, yForce/tractionRatio, zForce/tractionRatio), playerBall.CurrentState.Position);
		}
		
		//if ball has traction with another object
		private function hasTraction():Boolean
		{
			return JSphere(playerBall).Collisions.length > 0;
		}
		
		private function getCurrentSpeed():Number
		{
			return distanceBetweenTwoJNumber3D(playerBall.CurrentState.Position, playerBall.OldState.Position);
		}
		
		private function distanceBetweenTwoJNumber3D(firstJNumber3D:JNumber3D,secondJNumber3D:JNumber3D):Number
		{
			var x :Number = firstJNumber3D.x - secondJNumber3D.x;
			var y :Number = firstJNumber3D.y - secondJNumber3D.y;
			var z :Number = firstJNumber3D.z - secondJNumber3D.z;
	
			return Math.sqrt( x*x + y*y + z*z );
		}
		
		private function checkAllPhysicsObjects():void
		{
			var markerCounter:int = 0;
			var advancedPhysicsObj:AdvancedPhysicsObject;
			//for each (var advancedPhysicsObj:AdvancedPhysicsObject in physicsObjArray)
			for (var counter:int = 0; counter < physicsObjArray.length; counter++)
			{
				advancedPhysicsObj = physicsObjArray[counter];
				if (advancedPhysicsObj.isMarker || advancedPhysicsObj.isResettable)
				{
					if (advancedPhysicsObj.physicsObj.CurrentState.Position.y < WORLD_BOTTOM)
					{
						if (advancedPhysicsObj.isResettable)
						{
							advancedPhysicsObj.physicsObj.MoveTo(convertNumber3DToJNumber3D(advancedPhysicsObj.orginPosition),JMatrix3D.IDENTITY);
						}
						else if (advancedPhysicsObj.isMarker)
						{
							PhysicsSystem.getInstance().RemoveBody(advancedPhysicsObj.physicsObj);
							scene.removeChild(advancedPhysicsObj.physicsObj.BodySkin);
							physicsObjArray.splice(physicsObjArray.indexOf(advancedPhysicsObj), 1); //remove from array
							counter--;
							advancedPhysicsObj = null;
						}
					}
					else if (advancedPhysicsObj.isMarker)
					{
						markerCounter++;
					}
					
				}
			}
			
			remainingMarkers = markerCounter;
		}
		
		private function convertJNumber3DToNumber3D(valJNumber3D:JNumber3D):Number3D
		{
			return new Number3D(valJNumber3D.x, valJNumber3D.y, valJNumber3D.z);
		}
		
		private function convertNumber3DToJNumber3D(valNumber3D:Number3D):JNumber3D
		{
			return new JNumber3D(valNumber3D.x, valNumber3D.y, valNumber3D.z);
		}
		
	}//end class BallPhysics
}