// JigLibMotionSpringCamera3D Version 3 - Using Papervision 2.0 Beta - Revision 876, JibLibFlash - Revision 33, Caurina Tweener - Revision 404
//		JigLibMotionSpringCamera3D extends SpringCamera3D(http://agit8.turbulent.ca). It is a camera that follows the motion of JigLibFlash RigidBody object regardless on which way it is facing
//		Currently this camera only supports rotation on the Y axis by design
//
// released under MIT License
// http://www.opensource.org/licenses/mit-license.php
//
//@author		Devin Reimer
//@version		1.0.0
//@website 		http://www.permaven.com, http://blog.almostlogical.com

//Copyright (c) 2009 Devin Reimer
/*
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.cameras
{
	import org.papervision3d.core.math.*;
	import org.papervision3d.objects.DisplayObject3D;  
	import org.papervision3d.cameras.SpringCamera3D;
	import jiglib.physics.RigidBody;
	import jiglib.math.JNumber3D;

	/**
	 * JigLibMotionSpringCamera3D extends SpringCamera3D(http://agit8.turbulent.ca). It is a camera that follows the motion of JigLibFlash RigidBody object regardless on which way it is facing
	 * Currently this camera only supports rotation on the Y axis
	 * </p>
	 * 
	 * @author Devin Reimer - blog.almostlogical.com
	 */ 
	public class JigLibMotionSpringCamera3D extends SpringCamera3D
	{
		private var _jigLibObjTarget:RigidBody; //the object which the camera will follow the direction of it's motion, regardless on what way it is facing
												//if null acts like normal SpringCamera3D
		
		private var _reverseDeadZone:Number = 20; //the angle which when reversing will not cause the camera to spin around (note: reverseDeadZone*2 = total deadzone angle)
		private var _minimumTurningSpeed:Number = 1; //the minimum speed at which turning will occur
		private var _angleCorrectionSpeed:Number = 20.0; // the speed in which the camera will correct it's rotation, this is dependant on current speed
														 //  so it is recommended to be about twice that of maximum speed

		public function get jigLibObjTarget():RigidBody { return _jigLibObjTarget; }
		public function set jigLibObjTarget(val:RigidBody):void  {  _jigLibObjTarget = val; target = new DisplayObject3D(); }
		
		public function get reverseDeadZone():Number { return _reverseDeadZone; }
		public function set reverseDeadZone(val:Number) { _reverseDeadZone = val; }
		
		public function get minimumTurningSpeed():Number { return _minimumTurningSpeed; }
		public function set minimumTurningSpeed(val:Number) { _minimumTurningSpeed = val; }
		
		public function get angleCorrectionSpeed():Number { return _angleCorrectionSpeed; }
		public function set angleCorrectionSpeed(val:Number) { _angleCorrectionSpeed = val; }

		/**
		 * Constructor.
		 * 
		 * @param   fov     This value is the vertical Field Of View (FOV) in degrees.
		 * @param   near    Distance to the near clipping plane.
		 * @param   far     Distance to the far clipping plane.
		 * @param   useCulling      Boolean indicating whether to use frustum culling. When true all objects outside the view will be culled.
		 * @param   useProjection   Boolean indicating whether to use a projection matrix for perspective.
		 */  
		public function JigLibMotionSpringCamera3D(fov:Number = 60, near:Number = 10, far:Number = 5000, useCulling:Boolean = false, useProjection:Boolean = false)
		{
			super(fov, near, far, useCulling, useProjection);
		}
		
		//public function updateCameraPositionAndRotationBasedOnMotion():void
		public override function transformView(transform:Matrix3D = null):void
		{
			if (jigLibObjTarget != null)
			{
				updatePositionAndRotation();
			}
			
			super.transformView(transform);
		}
		
		//updates teh position and rotationY of the camera based on motion of the object
		private function updatePositionAndRotation():void
		{
			var angleOfMotion:Number = getPlayerAngleOfMotionInPV3DAngleSystem();
			var angleDistance:Number; //the shortest angular distance between two points (+ and -)
			var angleDirection:int = 0; // +1, 0, -1
			var angleToMove:Number;
			var currentSpeed:int = getRigidBodyCurrentSpeed();
			var actualAngleCorrectionSpeed:Number;
			
			target.rotationY = standardizePapervisionAngle(target.rotationY);
		
			angleDistance = getShortestDistanceBetweenTwoAngles(target.rotationY, angleOfMotion);
			
			if (angleDistance < 180 - reverseDeadZone && angleDistance>0) //if not in reverse dead zone and is a change in angle
			{
				angleDirection = getTurningDirection(target.rotationY, angleOfMotion);
			}
			else //in reverse dead zone or  not turning
			{
				angleDirection = 0;
			}
			
			if (currentSpeed >= minimumTurningSpeed) //only spin the the camera once minimum speed is achieved, this way it is not jerky as very slow speeds
			{
				actualAngleCorrectionSpeed = angleCorrectionSpeed - currentSpeed;
				if (actualAngleCorrectionSpeed < 1)  // actualAngleCorrectSpeed must be at least 1
				{
					actualAngleCorrectionSpeed = 1;
				}
				angleToMove = angleDistance / actualAngleCorrectionSpeed;  //the change in angle this time
				target.rotationY += (angleToMove * angleDirection);  //multiplies direction so angle is correct and sets new target rotationY
			}

			target.position = convertJNumber3DToNumber3D(jigLibObjTarget.CurrentState.Position); //update camera position
		}
		
		//get the shortest distance to rotate between two angles
		private function getShortestDistanceBetweenTwoAngles(angle1:Number, angle2:Number):Number
		{
			var shortestAngle:Number = (Math.abs(angle1 - angle2)) % 360;
			
			if (shortestAngle > 180)
			{
				shortestAngle = 360 - shortestAngle;
			}
			
			return shortestAngle;
		}
		
		//0-no turn, 1 - positive turn, -1 - negative turn
		//note: this function as good as it could be a better algorithm exists for this
		private function getTurningDirection(angle1:Number, angle2:Number):int
		{
			var differenceBetweenAngles:Number = angle1 - angle2;
			var angleDirection:int = 0;
			
			if ((differenceBetweenAngles > 0 && differenceBetweenAngles <= 180) || differenceBetweenAngles < -180)
			{	//move in negative direction
				angleDirection = -1;
			}
			else if ((differenceBetweenAngles < 0 && differenceBetweenAngles >= -180) || differenceBetweenAngles > 180)
			{	//move in positive direction
				angleDirection = 1;	
			}
			
			return angleDirection;
		}
		
		//as the rotation angles in papervision are different from standard angle system, covert
		// also only calculate angle on x and z axis
		private function getPlayerAngleOfMotionInPV3DAngleSystem():Number
		{
			var newJNumber:JNumber3D = jigLibObjTarget.CurrentState.Position;
			var oldJNumber:JNumber3D = jigLibObjTarget.OldState.Position;
			var dx:Number = (newJNumber.x - oldJNumber.x);
			var dz:Number = (newJNumber.z - oldJNumber.z);
			var angleRadians:Number = Math.atan2(dz,dx);
			var angleDegrees:Number = angleRadians * 180 / Math.PI;
			
			//better formula for this calculation
			angleDegrees -= 360 + 90;
			angleDegrees *= -1;
			angleDegrees %= 360;
			
			return angleDegrees;
		}
		
		//makes angle a position number between 0 - 359
		private function standardizePapervisionAngle(angle:Number):Number
		{
			angle = angle % 360;
			if (angle < 0) { angle += 360; }
			
			return angle;
		}
		
		private function getRigidBodyCurrentSpeed():Number
		{
			return distanceBetweenTwoJNumber3D(jigLibObjTarget.CurrentState.Position, jigLibObjTarget.OldState.Position);
		}
		
		private function distanceBetweenTwoJNumber3D(firstJNumber3D:JNumber3D,secondJNumber3D:JNumber3D):Number
		{
			var x :Number = firstJNumber3D.x - secondJNumber3D.x;
			var y :Number = firstJNumber3D.y - secondJNumber3D.y;
			var z :Number = firstJNumber3D.z - secondJNumber3D.z;
	
			return Math.sqrt( x*x + y*y + z*z );
		}
		
		private function convertJNumber3DToNumber3D(valJNumber3D:JNumber3D):Number3D
		{
			return new Number3D(valJNumber3D.x, valJNumber3D.y, valJNumber3D.z);
		}
		
	}//end class JigLibMotionSpringCamera3D
}
